from boa.interop.System.Runtime import Log
from boa.interop.System.Runtime import Notify
from boa.interop.System.Runtime import Serialize
from boa.interop.System.Runtime import Deserialize
from boa.interop.System.Storage import Put
from boa.interop.System.Storage import Get
from boa.interop.System.Storage import GetContext

ctx = GetContext()


def Main(operation ,args):
    if operation == 'save':
        return save()
    if operation == 'get':
        return get()
    return False

def save():
    #a = [1, 2, 3, 4]
    c = {'a':1,'b':2}

    b = Serialize(c)

    Put(ctx,'test',b)
    Notify('saved')
    return True

def get():
    b = Get(ctx, 'test')
    a = Deserialize(b)
    Log(a['a'])
    Notify(a['b'])



def Base58ToAddress(arg):
    """

    :param arg:
    """
    pass

def AddressToBase58(arg):
    """

    :param arg:
    """
    pass

def GetCurrentBlockHash():
    """
    return current block hash
    """
    pass





from boa.interop.Neo.App import RegisterAppCall

OEP4Contract = RegisterAppCall('1b86175747d0c1490d5b48b20b10584e378ed947', 'operation', 'args')


def Main(operation, args):
    if operation == 'transfer':
        return OEP4Contract('transfer', args)

    return False





def GetHeight():
    """

    """
    pass


def GetHeader(height_or_hash):
    """

    :param height_or_hash:
    """
    pass


def GetBlock(height_or_hash):
    """

    :param height_or_hash:
    """
    pass


def GetTransactionByHash(hash):
    """

    :param hash:
    """
    pass


def GetContract(script_hash):
    """

    :param script_hash:
    """
    pass

def GetTransactionHeight(hash):
    """
    :para hash:   
    """
    pass



'''
cryptoMessage
'''

from boa.interop.System.Runtime import CheckWitness
from boa.interop.System.Runtime import GetTime, Notify
from boa.interop.System.Runtime import Serialize
from boa.interop.System.Storage import Put, Get, GetContext
from boa.builtins import concat
from boa.builtins import range

ctx = GetContext()
messageboxPrefix = 'MESSAGEBOX'
messageCountPrefix = 'MESSAGECOUNT'
pubkeyPrefix = 'PUBKEY'

def Main(operation, args):
    if operation == 'sendMessage':
        if len(args) != 4:
            return False
        from_addr = args[0]
        to_addr = args[1]
        encrypt = args[2]
        message = args[3]
        return sendMessage(from_addr, to_addr, encrypt, message)

    if operation == 'getMessage':
        if len(args) != 2:
            return False
        address = args[0]
        number = args[1]
        return getMessage(address, number)

    if operation == 'getMessageCount':
        if len(args) != 1:
            return False
        address = args[0]
        return getMessageCount(address)

    if operation == 'register':
        if len(args) != 2:
            return False
        address = args[0]
        pubkey = args[1]
        return register(address, pubkey)

    if operation == 'getPubkey':
        if len(args) != 1:
            return False
        address = args[0]
        return getPubkey(address)

    if operation == 'getRangeMessage':
        if len(args) != 3:
            return False
        address = args[0]
        start = args[1]
        end = args[2]
        return getRangeMessage(address,start,end)


    return False


def sendMessage(from_addr, to_addr, encrypt, message):
    '''
    send message to some address
    :param from_addr:
    :param to_addr:
    :param encrypt:
    :param message:
    :return:
    '''

    if CheckWitness(from_addr) == False:
        return False

    timestamp = GetTime()

    msg = {'FROM': from_addr, 'ENCRYPT': encrypt, 'MESSAGE': message, 'TIMESTAMP': timestamp}

    countkey = concatkey(messageCountPrefix, to_addr)
    count = Get(ctx, countkey)

    Put(ctx, concatkey(concatkey(messageboxPrefix, to_addr), count + 1), Serialize(msg))
    Put(ctx, countkey, count + 1)

    Notify(['sendMessage', from_addr, to_addr, count + 1 ])
    return True


def getMessage(address, number):
    key = concatkey(concatkey(messageboxPrefix, address), number)
    content = Get(ctx, key)
    return content


def getMessageCount(address):
    key = concatkey(messageCountPrefix, address)
    return Get(ctx, key)


def register(address, pubkey):

    if CheckWitness(address) == False:
        return False

    key = concatkey(pubkeyPrefix, address)
    if not Get(ctx, key):
        Put(ctx, key, pubkey)
        Notify(['register', address, pubkey])
    return True


def getRangeMessage(address, start, end):

    count = getMessageCount(address)

    if start < 1:
        start = 1
    if start > count:
        return None
    if end > count:
        end = count

    res = []
    for idex in range(start, end + 1):
        content = getMessage(address, idex)
        res.append(content)
    return Serialize(res)


def getPubkey(address):
    key = concatkey(pubkeyPrefix, address)
    return Get(ctx, key)


def concatkey(str1, str2):
    return concat(concat(str1, '_'), str2)







class Contract:

    @property
    def Script(self):
        """

        :return:
        """
        return GetScript(self)

    @property
    def StorageContext(self):
        """

        :return:
        """
        return GetStorageContext(self)


def GetStorageContext(contract):
    """

    :param contract:
    """
    pass

def Destroy():
    """

    :param contract:
    """
    pass



"""
An Example of OEP-4
"""
from boa.interop.System.Storage import GetContext, Get, Put, Delete
from boa.interop.System.Runtime import Notify, CheckWitness
from boa.builtins import concat, ToScriptHash
ctx = GetContext()

NAME = 'TokenName'
SYMBOL = 'Symbol'
DECIMAL = 8
FACTOR = 100000000
OWNER = ToScriptHash("AQf4Mzu1YJrhz9f3aRkkwSm9n3qhXGSh4p")
# OWNER = bytearray(b'\x61\x6f\x2a\x4a\x38\x39\x6f\xf2\x03\xea\x01\xe6\xc0\x70\xae\x42\x1b\xb8\xce\x2d')
TOTAL_AMOUNT = 1000000000
BALANCE_PREFIX = bytearray(b'\x01')
APPROVE_PREFIX = bytearray(b'\x02')

SUPPLY_KEY = 'totalSupply'


def Main(operation, args):
    """
    :param operation:
    :param args:
    :return:
    """
    # 'init' has to be invokded first after deploying the contract to store the necessary and important info in the blockchain
    if operation == 'init':
        return init()
    if operation == 'name':
        return name()
    if operation == 'symbol':
        return symbol()
    if operation == 'decimals':
        return decimals()
    if operation == 'totalSupply':
        return totalSupply()
    if operation == 'balanceOf':
        if len(args) != 1:
            return False
        acct = args[0]
        return balanceOf(acct)
    if operation == 'transfer':
        if len(args) != 3:
            return False
        else:
            from_acct = args[0]
            to_acct = args[1]
            amount = args[2]
            return transfer(from_acct,to_acct,amount)
    if operation == 'transferMulti':
        return transferMulti(args)
    if operation == 'transferFrom':
        if len(args) != 4:
            return False
        spender = args[0]
        from_acct = args[1]
        to_acct = args[2]
        amount = args[3]
        return transferFrom(spender,from_acct,to_acct,amount)
    if operation == 'approve':
        if len(args) != 3:
            return False
        owner  = args[0]
        spender = args[1]
        amount = args[2]
        return approve(owner,spender,amount)
    if operation == 'allowance':
        if len(args) != 2:
            return False;
        owner = args[0]
        spender = args[1]
        return allowance(owner,spender)


def init():
    """
    initialize the contract, put some important info into the storage in the blockchain
    :return:
    """
    if Get(ctx,SUPPLY_KEY):
        Notify('Already initialized!')
        return False
    else:
        total = TOTAL_AMOUNT * FACTOR
        Put(ctx,SUPPLY_KEY,total)
        Put(ctx,concat(BALANCE_PREFIX,OWNER),total)
        Notify(['transfer', '', OWNER, total])
        return True

def name():
    """
    :return: name of the token
    """
    return NAME


def symbol():
    """
    :return: symbol of the token
    """
    return SYMBOL


def decimals():
    """
    :return: the decimals of the token
    """
    return DECIMAL


def totalSupply():
    """
    :return: the total supply of the token
    """
    return Get(ctx, SUPPLY_KEY)


def balanceOf(account):
    """
    :param account:
    :return: the token balance of account
    """
    return Get(ctx,concat(BALANCE_PREFIX,account))


def transfer(from_acct,to_acct,amount):
    """
    Transfer amount of tokens from from_acct to to_acct
    :param from_acct: the account from which the amount of tokens will be transferred
    :param to_acct: the account to which the amount of tokens will be transferred
    :param amount: the amount of the tokens to be transferred
    :return: True means success, False or raising exception means failure.
    """
    if from_acct == to_acct or amount == 0:
        return True
    if CheckWitness(from_acct) == False:
        return False
    if len(to_acct) != 20 or len(from_acct) != 20:
        return False
    if amount < 0:
        return False
    fromKey = concat(BALANCE_PREFIX,from_acct)
    fromBalance = Get(ctx,fromKey)
    if amount > fromBalance:
        return False
    if amount == fromBalance:
        Delete(ctx,fromKey)
    else:
        Put(ctx,fromKey,fromBalance - amount)

    toKey = concat(BALANCE_PREFIX,to_acct)
    toBalance = Get(ctx,toKey)
    Put(ctx,toKey,toBalance + amount)

    Notify(['transfer',from_acct,to_acct,amount])

    return True


def transferMulti(args):
    """
    :param args: the parameter is an array, containing element like [from, to, amount]
    :return: True means success, False or raising exception means failure.
    """
    for p in args:
        if len(p) != 3:
            # return False
            raise Exception("transferMulti params error.")
        if transfer(p[0], p[1], p[2]) == False:
            # return False # this is wrong since the previous transaction will be successful
            raise Exception("transferMulti failed.")
    return True


def approve(owner,spender,amount):
    """
    owner allow spender to spend amount of token from owner account
    Note here, the amount should be less than the balance of owner right now.
    :param owner:
    :param spender:
    :param amount:
    :return: True means success, False or raising exception means failure.
    """
    if amount < 0:
        return False
    if amount > balanceOf(owner):
        return False
    if CheckWitness(owner) == False:
        return False
    if len(spender) != 20 or len(owner) != 20:
        raise Exception("approve params failed")
    key = concat(concat(APPROVE_PREFIX,owner),spender)
    Put(ctx, key, amount)
    Notify(['approval', owner, spender, amount])
    return True


def transferFrom(spender,from_acct,to_acct,amount):
    """
    spender spend amount of token on the behalf of from_acct, spender makes a transaction of amount of tokens
    from from_acct to to_acct
    :param spender:
    :param from_acct:
    :param to_acct:
    :param amount:
    :return:
    """
    if CheckWitness(spender) == False:
        return False
    if len(spender) != 20 or len(from_acct) != 20 or len(to_acct) != 20:
        raise Exception("transferFrom params failed")
    if amount < 0:
        return False
    fromKey = concat(BALANCE_PREFIX, from_acct)
    fromBalance = Get(ctx, fromKey)
    if amount > fromBalance:
        return False

    approveKey = concat(concat(APPROVE_PREFIX,from_acct),spender)
    approvedAmount = Get(ctx,approveKey)
    toKey = concat(BALANCE_PREFIX,to_acct)
    if amount > approvedAmount:
        return False
    elif amount == approvedAmount:
        Delete(ctx,approveKey)
        Delete(ctx, fromKey)
    else:
        Put(ctx,approveKey,approvedAmount - amount)
        Put(ctx, fromKey, fromBalance - amount)

    toBalance = Get(ctx, toKey)    
    Put(ctx, toKey, toBalance + amount)

    Notify(['transfer',spender, from_acct,to_acct,amount])

    return True


def allowance(owner,spender):
    """
    check how many token the spender is allowed to spend from owner account
    :param owner: token owner
    :param spender:  token spender
    :return: the allowed amount of tokens
    """
    allowanceKey = concat(concat(APPROVE_PREFIX,owner),spender)
    return Get(ctx,allowanceKey)




from boa.interop.System.Runtime import *
from boa.interop.Ontology.Native import *
from boa.builtins import state

contractAddress = bytearray(b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01')


def Main(operation, args):
    if operation == 'transfer':
        fromacct = args[0]
        toacct = args[1]
        amount = args[2]
        return transferONT(fromacct,toacct,amount)


    return False


def transferONT(fromacct,toacct,amount):
    """
    transfer ONT
    :param fromacct:
    :param toacct:
    :param amount:
    :return:
    """
    if CheckWitness(fromacct):

        param = makeState(fromacct, toacct, amount)
        res = Invoke(1, contractAddress, 'transfer', [param])
        Notify(res)

        if res and res == b'\x01':
            Notify('transfer succeed')
            return True
        else:
            Notify('transfer failed')

            return False

    else:
        Notify('checkWitness failed')
        return False


def makeState(fromacct,toacct,amount):
    return state(fromacct, toacct, amount)


from bytecode import UNSET, Label, Instr, Bytecode, BasicBlock, ControlFlowGraph
from boa.code import pyop
import glob
import importlib
import hashlib
import base58


class BlockType():
    MAKE_FUNCTION = 0
    CALL_FUNCTION = 1
    MAKE_CLASS = 2
    IMPORT_ITEM = 3
    MODULE_VAR = 4
    DOC_STRING = 5
    LOAD_CONST = 6
    ACTION_REG = 7
    APPCALL_REG = 8
    UNKNOWN = 9


def get_block_type(block):

    for instr in block:
        if instr.opcode == pyop.LOAD_NAME and instr.arg == 'RegisterAction':
            return BlockType.ACTION_REG
        elif instr.opcode == pyop.LOAD_NAME and instr.arg == 'RegisterAppCall':
            return BlockType.APPCALL_REG
        elif instr.opcode in [pyop.IMPORT_FROM, pyop.IMPORT_NAME, pyop.IMPORT_STAR]:
            return BlockType.IMPORT_ITEM
        elif instr.opcode == pyop.MAKE_FUNCTION:
            return BlockType.MAKE_FUNCTION
        elif instr.opcode == pyop.LOAD_BUILD_CLASS:
            return BlockType.MAKE_CLASS
        elif instr.opcode == pyop.CALL_FUNCTION:
            return BlockType.CALL_FUNCTION

    return BlockType.UNKNOWN

class Digest(object):
    @staticmethod
    def __sha256(msg: bytes, is_hex: bool = False):
        m = hashlib.sha256()
        m.update(msg)
        if is_hex:
            return m.hexdigest()
        else:
            return m.digest()

    @staticmethod
    def ripemd160(msg: bytes, is_hex: bool = False):
        h = hashlib.new('ripemd160')
        h.update(msg)
        if is_hex:
            return h.hexdigest()
        else:
            return h.digest()

    @staticmethod
    def sha256(msg: bytes, offset: int = 0, length: int = 0, is_hex: bool = False):
        if offset != 0 and len(msg) > offset + length:
            msg = msg[offset:offset + length]
        return Digest.__sha256(msg, is_hex)

    @staticmethod
    def hash256(msg: bytes, is_hex: bool = False) -> bytes or str:
        digest = Digest.sha256(Digest.sha256(msg), is_hex=is_hex)
        return digest

    @staticmethod
    def hash160(msg: bytes, is_hex: bool = False) -> bytes or str:
        digest = Digest.ripemd160(Digest.__sha256(msg), is_hex)
        return digest

class Address(object):
    __COIN_VERSION = b'\x17'

    def __init__(self, value: bytes):
        self.ZERO = value  # 20 bytes

    def to_array(self):
        return self.ZERO

    def b58encode(self):
        script_builder = Address.__COIN_VERSION + self.ZERO
        c256 = Digest.hash256(script_builder)[0:4]
        out_byte_array = script_builder + bytearray(c256)
        return base58.b58encode(bytes(out_byte_array)).decode('utf-8')

    @staticmethod
    def b58decode(address: str):
        data = base58.b58decode(address)
        if len(data) != 25:
            raise Exception("ERROR:Wrong data len")
        if data[0] != int.from_bytes(Address.__COIN_VERSION, "little"):
            raise Exception("ERROR: Adrress VERSION wrong")
        checksum = Digest.hash256(data[0:21])
        if data[21:25] != checksum[0:4]:
            raise Exception("ERROR: checksum ERROR")
        return Address(data[1:21])



"""
An Example on domain auction
"""
from boa.interop.System.Storage import *
from boa.interop.System.Runtime import *
from boa.interop.System.ExecutionEngine import *
from boa.builtins import concat
from boa.interop.Ontology.Native import *
from boa.builtins import state

ctx = GetContext()
selfAddr = GetExecutingScriptHash()

#ONT native contract address
contractAddress = bytearray(b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01')


def Main(operation, args):

    if operation == 'register':
        acct = args[0]
        url = args[1]
        if CheckWitness(acct):
            return register(acct,url)
        Notify('CheckWitness failed!')
        return False
    
    if operation == 'sell':
        acct = args[0]
        url = args[1]
        price = args[2]
        if CheckWitness(acct):
            return sell(acct,url,price)
        Notify('CheckWitness failed!')
        return False

    if operation == 'query':
        url = args[0]
        return query(url)

    if operation == 'buy':
        acct = args[0]
        url = args[1]
        price = args[2]
        if CheckWitness(acct):
            return buy(acct,url,price)
        Notify('CheckWitness failed!')
        return False

    if operation == 'done':
        acct = args[0]
        url = args[1]
        if CheckWitness(acct):
            return done(acct,url)
        Notify('CheckWitness failed!')
        return False

    if operation == 'transfer':
        fromacct = args[0]
        amount = args[1]
        return transferONT(fromacct,selfAddr,amount)

    Notify('Not a supported operation!')
    return True


def register(account,domain):
    """
    register an domain for account
    :param account:
    :param domain:
    :return:
    """
    if not Get(ctx,domain):
        Put(ctx,domain,account)
        Notify('register succeed!')
        return True
    Notify('already registered!')
    return False


def sell(account,url, price):
    """
    sell the domain at a price
    :param account:
    :param url:
    :param price:
    :return:
    """
    owner = Get(ctx,url)
    if owner == account :
        Put(ctx,concat('Original_Owner_',url),account)
        Put(ctx,concat('Price_',url),price)
        Put(ctx,url,selfAddr)
        Notify('sell succeed!')
        return True
    Notify('Not a owner')
    return False


def query(url):
    """
    query a domain owner
    :param url:
    :return:
    """
    owner = Get(ctx,url)
    Notify(concat('owner is ',owner))
    return owner


def buy(acct,url,price):
    """
    buy a domain a price
    :param acct:
    :param url:
    :param price:
    :return:
    """
    owner = Get(ctx,url)
    if owner != selfAddr:
        Notify("url not in sale!")
        return False

    prevBuyer = Get(ctx,concat('TP_',url))
    currentPrice = Get(ctx,concat('Price_',url))
    #no buyer before case
    if not prevBuyer:
        if price >= currentPrice:
            if transferONT(acct, selfAddr, price):
                Put(ctx,concat('TP_',url),acct)
                if price > currentPrice:
                    Put(ctx,concat('Price_',url),price)
                Notify('buy succeed!')
                return True
            else:
                Notify('Transfer Failed')
                return False
        Notify('Price is lower than current price')
        return False
    # has buyer before case
    if price <= currentPrice:
        Notify('Price is lower than current price')
        return False
    if transferONT(selfAddr,acct,currentPrice):
        Put(ctx,concat('TP_',url),acct)
        Put(ctx,concat('Price_',url),price)
        Notify('refund succeed!')
        return True
    else:
        Notify('refund failed')
        return False


def transferONT(fromacct,toacct,amount):
    """
    transfer ONT
    :param fromacct:
    :param toacct:
    :param amount:
    :return:
    """
    if CheckWitness(fromacct):
        param = makeState(fromacct, toacct, amount)
        res = Invoke(1,contractAddress,'transfer',[param])
        Notify(res)

        if res and res == b'\x01':
            Notify('transfer succeed')
            return True
        else:
            Notify('transfer failed')

            return False

    else:
        Notify('checkWitness failed')
        return False


def done(acct,url):
    """
    finish the domain auction
    :param acct:
    :param url:
    :return:
    """
    currentOwner = Get(ctx, url)
    if currentOwner != selfAddr:
        Notify('not in sell')
        return False
    preOwner = Get(ctx,concat('Original_Owner_', url))
    if preOwner != acct:
        Notify('not owner')
        return False
    amount = Get(ctx, concat('Price_', url))
    param = makeState(selfAddr, acct, amount)
    res = Invoke(1, contractAddress, 'transfer', [param])
    if res and res == b'\x01':
        buyer = Get(ctx, concat('TP_',url))
        Put(ctx, url, buyer)

        Delete(ctx,concat('TP_', url))
        Delete(ctx,concat('Price_', url))
        Delete(ctx,concat('Original_Owner_', url))
        Notify('done succeed!')
        return True
    else:
        Notify('transfer failed')
        return False

def makeState(fromacct,toacct,amount):
    return state(fromacct, toacct, amount)